/***************************************************************************
                          account.cpp  -  description
                             -------------------
    begin                : Sat Aug 10 2002
    copyright            : (C) 2002 by Richard Garand
    email                : richard@garandnet.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdexcept>

#include <qtextstream.h>

#include <kdebug.h>
 
#include "account.h"
#include "kbudgetdoc.h"
#include "kbundoaction.h"
#include "kperiod.h"

Account::Account(KBudgetDoc* doc)
  : initialBalance(0), m_doc(doc)
{

}

Account::Account(KBudgetDoc* doc, int id, types type, QTextStream &stream, int streamVersion)
{
  m_doc = doc;
  m_id = id;
  stream >> initialBalance;
  m_type = (types)type;
  m_name = stream.readLine().stripWhiteSpace();
  dump();

  if ( streamVersion == 2 ) {
    int budgetValues, year, month;
    float value;
    stream >> budgetValues;
    for ( int i = 0; i < budgetValues; i++ ) {
      stream >> year >> month >> value;
      setBudgeted(QDate(year, month, 1), value);
    }
  }
}

Account::Account(KBudgetDoc* doc, int id, types type, QDomElement element)
  : m_type((types)type), m_doc(doc), m_id(id)
{
  bool ok;

  // read basic information
  initialBalance = element.attributeNode("initialbalance").value().toFloat(&ok);
  if ( !ok )
    throw std::runtime_error("invalid initial balance");

  m_name = element.attributeNode("name").value();

  // dump data
  dump();

  // read budget values
  unsigned int year, month;
  float value;
  QDomNodeList nodes = element.elementsByTagName("budgetentry");
  for ( unsigned int i = 0; i < nodes.count(); i++ ) {
    QDomElement e = nodes.item(i).toElement();
    
    year = e.attributeNode("year").value().toUInt(&ok);
    if ( !ok ) {
      kdWarning() << "Error reading budget entry: invalid year" << endl;
      continue;
    }

    month = e.attributeNode("month").value().toUInt(&ok);
    if ( !ok ) {
      kdWarning() << "Error reading budget entry: invalid month" << endl;
      continue;
    }

    value = e.attributeNode("amount").value().toFloat(&ok);
    if ( !ok ) {
      kdWarning() << "Error reading budget entry: invalid amount" << endl;
      continue;
    }

    setBudgeted(QDate(year, month, 1), value);
  }
}

Account::Account(KBudgetDoc* d, int i, types t, QString n, float ib)
{
  m_type = t;
  m_name = n;
  initialBalance = ib;
  m_doc = d;
  m_id = i;
  dump();
}

Account::Account(const Account& from)
{
  m_type = from.m_type;
  m_name = from.m_name;
  months = from.months;
  initialBalance = from.initialBalance;
  m_doc = from.m_doc;
  m_id = from.m_id;
}

Account::~Account()
{

}

void Account::serialize(QDomElement &parent)
{
  // add base information
  QDomElement e = parent.ownerDocument().createElement("account");
  e.setAttribute("id", m_id);
  e.setAttribute("type", (int)m_type);
  e.setAttribute("initialbalance", initialBalance);
  e.setAttribute("name", m_name);

  // add budget entries
  if ( months.size() ) {
    QMap<int,year_t>::const_iterator year;
    year_t::const_iterator month;
    for ( year = months.begin(); year != months.end(); year++ ) {
      for ( month = (*year).begin(); month != (*year).end(); month++ ) {
        if ( (*month).budgeted ) {
          QDomElement b = e.ownerDocument().createElement("budgetentry");
          b.setAttribute("year", year.key());
          b.setAttribute("month", month.key());
          b.setAttribute("amount", (*month).budgeted);
          e.appendChild(b);
        }
      }
    }
  }

  // append the element
  parent.appendChild( e );
}

int Account::id() const
{
  return m_id;
}

void Account::setType(types tp)
{
  m_type = tp;
}

Account::types Account::type() const
{
  return m_type;
}

void Account::setName(QString n)
{
  if ( m_doc->canAddUndo() )
    m_doc->addUndo(new KBAccountUndo(m_doc, KBAccountUndo::UA_RENAME, m_id));
  m_name = n;
  m_doc->signalAcctChanged(m_id);
}

QString Account::name() const
{
	return m_name;
}

void Account::addTransaction(int id)
{
  transactionUpdate(id, true);
}

void Account::removeTransaction(int id)
{
  transactionUpdate(id, false);
}

QValueList<int> Account::getTransactions(const KPeriod &period) const
{
  QValueList<int> list;

  if ( !months.size() )
    return list;

  QMap<int,year_t>::const_iterator yr;// = months.begin()
  year_t::const_iterator mn;
  for ( yr = months.begin(); yr != months.end(); yr++ ) {
    if ( !period.containsYear(yr.key()) )
      continue;
    for ( mn = (*yr).begin(); mn != (*yr).end(); mn++ ) {
      if ( !period.containsMonth(yr.key(), mn.key()) )
        continue;
      for ( QValueList<int>::const_iterator it = (*mn).transactions.begin(); it != (*mn).transactions.end(); it++ ) {
        const Transaction& tx = m_doc->getTransaction(*it);
        if ( period.type() != KPeriod::DateRange || period.contains(tx.date()) )
          addTransaction(list, tx);
      }
    }
  }

  return list;
}

/*float Account::getBalance() const
{
  if ( months.size() )
    return (*--(*--months.end()).end()).endBalance;
  else
    return initialBalance;
}*/

float Account::startingBalance() const
{
  return initialBalance;
}

/*float Account::startingBalance(const KPeriod &period) const
{
  QDate fd = period.firstDay();

  QMap<int,year_t>::const_iterator yr;
  year_t::const_iterator mn;
  for ( yr = months.begin(); yr != months.end(); yr++ ) {
    if ( fd.year() < yr.key() )
      continue;
    for ( mn = (*yr).begin(); mn != (*yr).end(); mn++ ) {
      if ( fd.month() < mn.key() )
        continue;
      return (*mn).startBalance;
    }
  }

  return 0;
}*/

void Account::setStartingBalance(float balance)
{
  if ( m_doc->canAddUndo() )
    m_doc->addUndo(new KBAccountUndo(m_doc, KBAccountUndo::UA_STBALANCE, m_id));

  float diff = balance - initialBalance;
  initialBalance = balance;
  QMap<int,year_t>::iterator yr = months.begin();
  year_t::iterator month = (*yr).begin();
  add(yr, month, diff, 0.0f);

  m_doc->signalAcctChanged(m_id);
}

float Account::endingBalance(QDate day) const
{
  QMap<int,Account::year_t>::const_iterator yr = months.find(day.year());
  if ( yr != months.end() && (*yr).find(day.month()) != (*yr).end() ) {
    // we have data for this month; take initial balance and add all transactions
    // up to day
    const Month &m = months[day.year()][day.month()];
    float balance = m.startBalance;
    for ( QValueList<int>::const_iterator it = m.transactions.begin(); it != m.transactions.end(); it++ ) {
      Transaction& tr = m_doc->getTransaction(*it);
      if ( tr.date() <= day ) {
        if ( tr.from() == m_id )
          balance -= tr.value();
        else
          balance += tr.value();
      }
    }
    return balance;
  } else {
    const Month *m = monthBefore(day);
    if ( m )
      // we have no data for the given month, so we just take the ending balance
      // of m
      return m->endBalance;
  }
  
  // we have no data before or on day, so just take the initial balance
  return initialBalance;
}

/*float Account::change() const
{
  if ( m_type == AC_INCOME )
    return initialBalance - getBalance();
  else
    return getBalance() - initialBalance;
}

float Account::startingBudget(const KPeriod &period) const
{
  const Month* month = getFirstMonth(period);
  if ( month )
    return month->startBudget;
  return 0;
}

float Account::budget(QDate month) const
{
  if ( months.size() )
    if ( month.isNull() )
      return (*--(*--months.end()).end()).endBudget;
    else if ( months.find(month.year()) != months.end() && \
      months[month.year()].find(month.month()) != months[month.year()].end() )
      return months[month.year()][month.month()].endBudget;
    else
      return 0;
  else
    return 0;
}*/

float Account::budgeted(QDate month) const
{
  if ( months.size() )
    if ( month.isNull() )
      return (*--(*--months.end()).end()).budgeted;
    else if ( months.find(month.year()) != months.end() && \
      months[month.year()].find(month.month()) != months[month.year()].end() )
      return months[month.year()][month.month()].budgeted;
    else
      return 0;
  else
    return 0;
}

void Account::setBudgeted(QDate month, float value)
{
  QMap<int,year_t>::iterator yearIt;
  year_t::iterator monthIt;
  createMonth(month.year(), month.month(), yearIt, monthIt);

  float addBudget = value - (*monthIt).budgeted;
  if ( !addBudget )
    return;

  if ( m_doc->canAddUndo() )
    m_doc->addUndo(new KBAccountUndo(m_doc, KBAccountUndo::UA_BUDGET, m_id, month));

  (*monthIt).budgeted = value;
  (*monthIt).endBudget += addBudget;

  monthIt++;
  add(yearIt, monthIt, 0, addBudget);

  m_doc->signalAcctChanged(m_id);
}

Account Account::clone(QDate fromDate, QDate toDate) const
{
  Account newAcct = *this;
  newAcct.months.clear();

  QMap<int,year_t>::const_iterator year;
  year_t::const_iterator month;
  bool firstMonth = true;

  for ( year = months.begin(); year != months.end(); year++ )
    for ( month = (*year).begin(); month != (*year).end(); month++ )
      if ( isInRange(fromDate, toDate, QDate(year.key(), month.key(), 1)) ) {
        if ( firstMonth ) {
          newAcct.initialBalance = (*month).startBalance;
          firstMonth = false;
        }
        newAcct.months[year.key()][month.key()] = *month;
      }

  // no months were added to the clone, so we need to set the initial balance
  const Month* m = monthBefore(fromDate);
  if ( firstMonth )
    if ( m ) {
      newAcct.initialBalance = m->endBalance;
      newAcct.months[fromDate.year()][fromDate.month()].startBudget = m->endBudget;
      newAcct.months[fromDate.year()][fromDate.month()].endBudget = m->endBudget;
      newAcct.months[fromDate.year()][fromDate.month()].startBalance = newAcct.initialBalance;
      newAcct.months[fromDate.year()][fromDate.month()].endBalance = newAcct.initialBalance;
    } else
      newAcct.initialBalance = initialBalance;

  return newAcct;
}

QDate Account::firstMonth() const
{
  if ( months.size() )
    return QDate(months.begin().key(), (*months.begin()).begin().key(), 1);
  else
    return QDate::currentDate();
}

const Account::Month* Account::monthBefore(QDate date) const
{
  QMap<int,year_t>::const_iterator year;
  year_t::const_iterator month;
  year_t::const_iterator prevmonth;
  if ( months.size() ) {
    QDate firstMonth(months.begin().key(), (*months.begin()).begin().key(), 1);
    if ( firstMonth < date ) {
      for ( year = months.begin(); year != months.end(); year++ )
        for ( month = (*year).begin(); month != (*year).end(); month++ )
          if ( year.key() > date.year() || (year.key() == date.year() && month.key() > date.month()) )
            return &*prevmonth;
          else
            prevmonth = month;
      return &*--month;
    } // end of have previous month
  } // end of have months
  return 0;
}

bool Account::isInRange(QDate from, QDate to, QDate test) const
{
  if ( to.isValid() ) {
    if ( test.year() < from.year() || test.year() > to.year() )
      return false;
    if ( from.month() && test.year() == from.year() && test.month() < from.month() )
      return false;
    if ( to.month() && test.year() == to.year() && test.month() > to.month() )
      return false;
    return true;
  }
  if ( test.year() != from.year() )
    return false;
  if ( from.month() && test.month() != from.month() )
    return false;
  return true;
}

void Account::createMonth(int year, int month, QMap<int,year_t>::iterator &yearIt, year_t::iterator &monthIt)
{
  // if the month doesn't exist, create and initialize it
  if ( months.find(year) == months.end() || months[year].find(month) == months[year].end() ) {
    // create the month and set default values
    months[year][month].startBalance = initialBalance;
    months[year][month].endBalance = initialBalance;

    // set the iterators
    yearIt = months.find(year);
    monthIt = (*yearIt).find(month);

    // look for a previous month
    if ( monthIt != (*yearIt).begin() )
      monthIt--;
    else
      if ( yearIt != months.begin() ) {
        yearIt--;
        monthIt = --(*yearIt).end();
      } else
        // don't initialize month
        yearIt = months.end();

    // copy previous month's data
    if ( yearIt != months.end() ) {
      months[year][month].startBalance = (*monthIt).endBalance;
      months[year][month].endBalance = (*monthIt).endBalance;
      months[year][month].startBudget = (*monthIt).endBudget;
      months[year][month].endBudget = (*monthIt).endBudget;
    }
  }

  yearIt = months.find(year);
  monthIt = (*yearIt).find(month);
}

void Account::dump() const
{
//  kdDebug() << "Account " << m_id << ": " << initialBalance << " to " << getBalance() << endl;
}

Account& Account::operator= (const Account &from)
{
  m_type = from.m_type;
  m_name = from.m_name;
  months = from.months;
  initialBalance = from.initialBalance;
  m_doc = from.m_doc;
  m_id = from.m_id;

  return *this;
}

void Account::transactionUpdate(int id, bool adding)
{
  Transaction &tr = m_doc->getTransaction(id);
  QDate date = tr.date();

  QMap<int,year_t>::iterator year;
  year_t::iterator month;
  //if ( adding )
    createMonth(date.year(), date.month(), year, month);
  Month& mnth = months[date.year()][date.month()];

  float addBalance = tr.value();
  // if it's from, we want to subtract, not add
  addBalance *= m_id == tr.to() ? 1 : m_id == tr.from() ? -1 : 0;
  // if we're removing, negate the value
  if ( !adding )
    addBalance *= -1;
  mnth.endBalance += addBalance;

  float budgetChange = addBalance;
  budgetChange *= m_type == AC_EXPENSE ? -1 : 1;
  mnth.endBudget += budgetChange;

  if ( adding )
    mnth.transactions.push_back(id);
  else {
    mnth.transactions.remove(id);
    // clean up the map
    // TODO: re-enable this and fix it
/*    if ( !mnth.transactions.size() ) {
      months[date.year()].remove(date.month());
      if ( months[date.year()].keys().size() == 0 )
        months.remove(date.year());
    }*/
  }

  month++;
  add(year, month, addBalance, budgetChange);

  m_doc->signalAcctChanged(m_id);
}

void Account::add(QMap<int,year_t>::iterator year, year_t::iterator month, float balance, float budget)
{
  while ( year != months.end() ) {
    while ( month != (*year).end() ) {
      (*month).startBalance += balance;
      (*month).endBalance += balance;
      (*month).startBudget += budget;
      (*month).endBudget += budget;
      month++;
    }
    year++;
    if ( year != months.end() )
      month = (*year).begin();
  }
}

void Account::addTransaction(QValueList<int> &list, const Transaction &add) const
{
  // TODO: store the transactions in sorted lists to avoid this
  for ( QValueList<int>::iterator it = list.begin(); it != list.end(); it++ ) {
    if ( m_doc->getTransaction(*it).date() > add.date() ) {
      list.insert(it, add.id());
      return;
    }
  }

  list.push_back(add.id());
}

const Account::Month* Account::getFirstMonth(const KPeriod &period) const
{
  QDate fd = period.firstDay();

  QMap<int,year_t>::const_iterator yr;
  year_t::const_iterator mn;
  for ( yr = months.begin(); yr != months.end(); yr++ ) {
    if ( fd.year() < yr.key() )
      continue;
    for ( mn = (*yr).begin(); mn != (*yr).end(); mn++ ) {
      if ( fd.month() < mn.key() )
        continue;
      return &(*mn);
    }
  }
  return 0;
}


AccountAccessor::AccountAccessor()
  : acct(0)
{
}

AccountAccessor::AccountAccessor(const Account* a, const KPeriod& p)
  : acct(a), period(p)
{
  setMonth();
}

AccountAccessor::AccountAccessor(const AccountAccessor& from)
  : acct(from.acct), period(from.period), month(from.month)
{
}

void AccountAccessor::setPeriod(const KPeriod& to)
{
  period = to;
  setMonth();
}

int AccountAccessor::id() const
{
  if ( !acct )
    return -1;
  return acct->id();
}

Account::types AccountAccessor::type() const
{
  return acct->type();
}

QString AccountAccessor::name() const
{
  return acct->name();
}

float AccountAccessor::balance() const
{
  return month.endBalance;
}

float AccountAccessor::startingBalance() const
{
  return month.startBalance;
}

float AccountAccessor::change() const
{
  if ( acct->m_type == Account::AC_INCOME )
    return -(month.endBalance - month.startBalance);
  return month.endBalance - month.startBalance;
}

float AccountAccessor::startingBudget() const
{
  return month.startBudget;
}

float AccountAccessor::budget() const
{
  return month.endBudget;
}

float AccountAccessor::budgeted() const
{
  return month.budgeted;
}

AccountAccessor& AccountAccessor::operator= (const AccountAccessor& a)
{
  acct = a.acct;
  period = a.period;
  month = a.month;

  return *this;
}

void AccountAccessor::refresh()
{
  setMonth();
}

void AccountAccessor::setMonth()
{
  QDate fd = period.firstDay();

  QMap<int,Account::year_t>::const_iterator yr = acct->months.find(fd.year());
  if ( yr != acct->months.end() && (*yr).find(fd.month()) != (*yr).end() ) {
    month = acct->months[fd.year()][fd.month()];
    // if the period doesn't include the full month, adjust the balance
    if ( period.type() == KPeriod::DateRange ) {
      float balanceChange = 0;
      KBudgetDoc* doc = acct->m_doc;

      // check all the transactions
      for ( QValueList<int>::iterator it = month.transactions.begin(); it != month.transactions.end(); it++ ) {
        Transaction& tr = doc->getTransaction(*it);
        if ( tr.date() < period.firstDay() )
          if ( tr.from() == acct->m_id )
            balanceChange -= tr.value();
          else
            balanceChange += tr.value();
      }

      // adjust the balance
      month.startBalance += balanceChange;
      month.endBalance += balanceChange;
    }
  } else {
    const Account::Month* m = acct->monthBefore(fd);
    if ( m ) {
      // m is the last month before the period we want; start == m->end
      month = *m;
      month.startBalance = month.endBalance;
      month.startBudget = month.endBudget;
      month.budgeted = 0;
    } else {
      // there are no months before or in the period; create a blank one
      month = Account::Month();
      month.startBalance = acct->initialBalance;
      month.endBalance = month.startBalance;
    }
  }
  // don't need these
  month.transactions.clear();
}
