/***************************************************************************
                          account.h  -  description
                             -------------------
    begin                : Sat Aug 10 2002
    copyright            : (C) 2002 by Richard Garand
    email                : richard@garandnet.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ACCOUNT_H
#define ACCOUNT_H

#include <qstring.h>
#include <qvaluelist.h>
#include <qdatetime.h>
#include <qmap.h>
#include <qdom.h>

#include "kperiod.h"

class KBudgetDoc;
class QTextStream;
class Transaction;

/**Contains information and a list of transactions for an account
  *@author Richard Garand
  */

class Account
{
  public:
    friend class KBudgetDoc;
    friend class AccountAccessor;

    /** account types */
    enum types { AC_INCOME = 0, AC_EXPENSE = 1, AC_CASH = 2, AC_ALL };

    /** default constructor (if doc == 0, no other functions can be called) */
    Account(KBudgetDoc* doc = 0);
    /** reads data from a stream; version is the stream's format version */
    Account(KBudgetDoc* doc, int id, types type, QTextStream &stream, int version);
    /** reads data from an XML element */
    Account(KBudgetDoc* doc, int id, types type, QDomElement e);
    /** sets account type */
    Account(KBudgetDoc* doc, int id, types type, QString name, float initialBalance);
    /** copy constructor */
    Account(const Account& from);
    /** Account destructor */
    ~Account();

    /** Writes the data to an XML document as an element under parent. */
    void serialize(QDomElement &parent);

    /** return the account ID */
    int id() const;
    
    /** set the type of the account */
    void setType(types type);
    /** return the type of the account */
    types type() const;

    /** sets the name of the account */
    void setName(QString name);
    /** returns the name of the account */
    QString name() const;

    /** adds a transaction to the account */
    void addTransaction(int id);
    /** removes a transaction from the account */
    void removeTransaction(int id);

    /** return a sorted list of transactions in the selected periods (0 = any) */
//    QValueList<int> getTransactions(int year = 0, int month = 0) const;
    /** return sorted list of transactions in the selected period (includes start and end dates) */
//    QValueList<int> getTransactions(QDate start, QDate end) const;
    /** return sorted list of transactions in the given period (includes start and end dates) */
    QValueList<int> getTransactions(const KPeriod &period) const;

    /** returns the starting balance */
    float startingBalance() const;
    /** sets the starting balance */
    void setStartingBalance(float balance);
    /** Returns the ending balance on the given day. */
    float endingBalance(QDate day) const;

    /** returns the amount budgeted for the specified month (Default is last month) */
    float budgeted(QDate date = QDate()) const;
    /** sets the amount budgeted for the specified month. Value for day can be anything */
    void setBudgeted(QDate month, float amount);

    /** returns a copy of the account. If month is set to 0, it will be ignored.
      * If toDate is set, select a range (including fromDate and toDate) */
    Account clone(QDate fromDate, QDate toDate = QDate()) const;

    /** returns the first day of the first month with transactions (defaults to current date).
      * This is just a temporary hack until we have better code for handling date ranges */
    QDate firstMonth() const;
    
    /** print debugging data */
    void dump() const;

    Account& operator= (const Account&);

  private:
    class Month {
      public:
        Month() : startBalance(0), endBalance(0), startBudget(0), endBudget(0), budgeted(0) {};
        float     startBalance;
        float     endBalance;
        QValueList<int> transactions;
        float     startBudget;
        float     endBudget;
        float     budgeted;
    };
    typedef QMap<int,Month> year_t;

    bool isInRange(QDate from, QDate to, QDate test) const;
    /** make sure a month is properly initialized */
    void createMonth(int year, int month, QMap<int,year_t>::iterator &yearIt, year_t::iterator &monthIt);
    /** return the last month before date or 0 */
    const Month* monthBefore(QDate date) const;
    /** add or remove a transaction */
    void transactionUpdate(int id, bool adding);
    /** add a certain amount to the start,end balance,budget for all months starting with the given iterators */
    void add(QMap<int,year_t>::iterator year, year_t::iterator month, float balance, float budget);
    void addTransaction(QValueList<int> &list, const Transaction &add) const;
    /** returns the first month in the given period */
    const Month* getFirstMonth(const KPeriod &period) const;

    /** type of account */
    types             m_type;
    /** name of account */
    QString           m_name;
    /** initial balance */
    float             initialBalance;
    /** list of transactions */
    QMap<int,year_t>  months;
    KBudgetDoc*       m_doc;
    int               m_id;
};

/** An account accessor interface. Initialized with an account and a period,
  * it's functions return information for the given period only. Invalid accessors
  * will return -1 from id(), but any other functions will have undefined results. */
class AccountAccessor
{
  public:
    AccountAccessor();
    AccountAccessor(const Account* acct, const KPeriod& period);
    AccountAccessor(const AccountAccessor& from);

    void setPeriod(const KPeriod& to);
    
    /** returns the account id or -1 if the accessor is invalid */
    int id() const;
    Account::types type() const;
    QString name() const;

    /** return the current balance */
    float balance() const;
    /** returns the starting balance */
    float startingBalance() const;
    /** return the difference between the starting balance and the current balance */
    float change() const;
    
    /** returns the initial budgeted amount */
    float startingBudget() const;
    /** returns the final budgeted amount */
    float budget() const;
    /** returns the amount budgeted for the specified month (Default is last month) */
    float budgeted() const;

    AccountAccessor& operator= (const AccountAccessor &);

    /** re-reads information from account */
    void refresh();

  private:
    // set the month
    void setMonth();

    const Account* acct;
    KPeriod period;
    Account::Month month;
};

#endif
