/***************************************************************************
                          kbchartwindow.cpp  -  description
                             -------------------
    begin                : Thu Oct 31 2002
    copyright            : (C) 2002 by Richard Garand
    email                : richard@garandnet.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
 
#include <config.h>

#include <qptrlist.h>
#include <qvbox.h>

#include <klocale.h>
#include <kaction.h>
#include <kmessagebox.h>
#include <kconfig.h>
#include <kdebug.h>

#ifdef HAVE_KDCHART
#include <KDChartWidget.h>
#include <KDChartAxisParams.h>
#include <KDChartTableBase.h>
#endif

#include "kbchartwindow.h"
#include "kbudgetdoc.h"
#include "kperiodselector.h"

QPtrList<KBChartWindow> KBChartWindow::instances;

KBChartWindow::KBChartWindow(KBudgetDoc *doc)
  : KMainWindow(0, "chart_window"), m_doc(doc), m_type(None), m_account(-1),
  m_acctType(Account::AC_EXPENSE), update(&KBChartWindow::updateNone)
{
#ifdef HAVE_KDCHART
  // configure chart
  m_params.setDataSubduedColors();
  m_params.setLegendSource(KDChartParams::LegendManual);
  m_params.setLegendTitleText(i18n("Accounts"));
#endif

  // initialize window
  instances.append(this);
  setCaption("Chart");

  QVBox *top = new QVBox(this);
  setCentralWidget(top);
  
  periodSelector = new KPeriodSelector(top);

#ifdef HAVE_KDCHART
  m_chartWidget = new KDChartWidget(&m_params, &m_data, top);
  top->setStretchFactor(m_chartWidget, 1);

  // segfault when window is made too small (!), this should fix it
  m_chartWidget->setMinimumSize(150, 100);
#endif

  // set up signals and slots
  connect(periodSelector, SIGNAL(periodChanged(KPeriod)), SLOT(slotPeriodChanged(KPeriod)));
  connect(doc, SIGNAL(accountChanged(int,KBudgetDoc::ChangeAction)), SLOT(slotAccountChanged(int)));
  connect(doc, SIGNAL(accountDeleted(int)), SLOT(slotAccountChanged(int)));
  connect(doc, SIGNAL(transactionChanged(int,KBudgetDoc::ChangeAction)), SLOT(slotTransactionChanged(int)));
  connect(doc, SIGNAL(transactionDeleted(int)), SLOT(slotTransactionChanged(int)));

  // initialize actions
  KStdAction::close(this, SLOT(close()), actionCollection())->setStatusText(i18n("Closes the window"));

  // initialize the GUI
  createGUI("kbchartwindowui.rc");

  // get configuration
  KConfig *conf = KGlobal::config();
  conf->setGroup("chartWindow_options");
  resize(500,400);    // using random numbers ripped from KNode :)
  applyMainWindowSettings(conf);
}

KBChartWindow::~KBChartWindow()
{
  instances.removeRef(this);
  KConfig *conf = KGlobal::config();
  conf->setGroup("chartWindow_options");
  saveMainWindowSettings(conf);
}

bool KBChartWindow::closeAllWindows()
{
  // TODO: limit this to the window that created this one
  QPtrList<KBChartWindow> list = instances;
  for (KBChartWindow* w = list.first(); w; w = list.next())
    w->close();
  return true;
}

void KBChartWindow::setPeriod(KPeriod period)
{
  periodSelector->setPeriod(period);
}

void KBChartWindow::setAccount(int account)
{
  m_account = account;
  (this->*update)();
}

void KBChartWindow::setAccountType(Account::types type)
{
  m_acctType = type;
  (this->*update)();
}

void KBChartWindow::setType(types type)
{
  m_type = type;
  switch ( m_type ) {
  case None: update = &KBChartWindow::updateNone; break;
  case ByAccount: update = &KBChartWindow::updateByAccount; break;
  case Balances: update = &KBChartWindow::updateBalances; break;
  }
  (this->*update)();
}

void KBChartWindow::updateNone()
{
#ifdef HAVE_KDCHART
  m_data.expand(0, 0);

  m_chartWidget->update();
#endif
}

void KBChartWindow::updateByAccount()
{
#ifdef HAVE_KDCHART
  // get the account list
  QValueVector<AccountAccessor> accts = m_doc->getAccounts(m_acctType, m_period);

  // resize the data
  m_data.expand(accts.size(), accts.size());

  // set up the chart parameters
  m_params.setChartType(KDChartParams::Pie);
//  m_params.setDataValuesPlacing(KDChartEnums::PosTopCenter, 0, 0, 0, 0);
  m_params.setPrintDataValues(true);
  m_params.setDataValuesCalc(0, 2);
  m_params.setDataValuesColors(&Qt::black);

  // add the account data to the chart
  for ( unsigned int i = 0; i < accts.size(); i++ ) {
    m_data.setCell(0, i, accts[i].change());
    m_params.setLegendText(i, accts[i].name());
  }

  m_chartWidget->update();
#endif
}

void KBChartWindow::updateBalances()
{
#ifdef HAVE_KDCHART
  // get account list
  QPtrList<Account> accts = m_doc->getAllAccounts(m_acctType);

  axisLabels.clear();

  // make data the right size
  m_data.expand(accts.count() + 1, m_period.length());

  // set up the chart
  m_params.setChartType(KDChartParams::Line);
  m_params.setLineWidth(2);
  m_params.setPrintDataValues(false);

  // set the data
  QDate dt = m_period.firstDay();
//  QDate start = dt;
  for ( int d = 0; d < m_period.length(); d++ ) {
    int an = 0;
    float total = 0;
    for ( Account *acct = accts.first(); acct; acct = accts.next() ) {
      float b = acct->endingBalance(dt);
      m_data.setCell(an, d, b);
      total += b;
      an++;
    }
    m_data.setCell(an, d, total);
    if ( dt.day() == 1 )
      axisLabels << QString("%1/%2").arg(dt.month()).arg(dt.year());
    else
      axisLabels << QString("%1").arg(dt.day());
    dt = dt.addDays(1);
  }

  m_params.setAxisLabelStringParams(KDChartAxisParams::AxisPosBottom, &axisLabels, &axisLabels);
/*  KDChartAxisParams ap(m_params.axisParams(KDChartAxisParams::AxisPosBottom));
  kdDebug() << "Setting start to " << start.addMonths(-1).toString(Qt::ISODate) << endl;
  ap.setAxisValueStart(KDChartData(0,start.addMonths(-1)));
  kdDebug() << "Setting end to " << dt.addMonths(1).toString(Qt::ISODate) << endl;
  ap.setAxisValueEnd(KDChartData(0,dt.addMonths(1)));
  m_params.setAxisParams(KDChartAxisParams::AxisPosBottom, ap);*/

  int an = 0;
  for ( Account* acct = accts.first(); acct; acct = accts.next() ) {
    m_params.setLegendText(an, acct->name());
    an++;
  }
  m_params.setLegendText(an, i18n("Total"));
  
  m_chartWidget->repaint();
#endif
}

void KBChartWindow::slotPeriodChanged(KPeriod period)
{
  m_period = period;
  (this->*update)();
}

void KBChartWindow::slotAccountChanged(int id)
{
  // NOTE: if we're only displaying one account, and it's deleted...
  if ( m_account >= 0 && m_account != id )
    return;
  // only abort here if the account exists and has the wrong type
  if ( m_acctType != Account::AC_ALL && m_doc->isValidAccount(id) && m_acctType != m_doc->getAccount(id).type() )
    return;
  (this->*update)();
}

void KBChartWindow::slotTransactionChanged(int id)
{
  // this just affects the acount(s), so...
  Transaction &tr = m_doc->getTransaction(id);
  slotAccountChanged(tr.from());
  slotAccountChanged(tr.to());
}
