/***************************************************************************
                          kbudgetdoc.cpp  -  description
                             -------------------
    begin                : Sat Aug 10 16:03:35 CST 2002
    copyright            : (C) 2002 by Richard Garand
    email                : richard@garandnet.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdexcept>

#include <qdir.h>
#include <qwidget.h>
#include <qdom.h>
#include <qtimer.h>

#include <klocale.h>
#include <kmessagebox.h>
#include <kio/job.h>
#include <kio/netaccess.h>
#include <ktempfile.h>
#include <ksavefile.h>
#include <kdebug.h>

#include "kbudgetdoc.h"
#include "kbudget.h"
#include "kbudgetview.h"
#include "kbundoaction.h"

using namespace std;

QList<KBudgetView> *KBudgetDoc::pViewList = 0L;

KBudgetDoc::KBudgetDoc(QWidget *parent, const char *name) : QObject(parent, name), nextAccount(0),
  nextTransaction(0), lastSavedPos(-1), inUndo(false), ugroup(0),
  blankAcct(this), blankTransaction(this), rTimer(new QTimer(this))
{
  if(!pViewList)
  {
    pViewList = new QList<KBudgetView>();
  }

  pViewList->setAutoDelete(true);

  undos.setAutoDelete(true);

  connect(rTimer, SIGNAL(timeout()), SLOT(slotCheckRecurrences()));
}

KBudgetDoc::~KBudgetDoc()
{
}

void KBudgetDoc::addView(KBudgetView *view)
{
  pViewList->append(view);
}

void KBudgetDoc::removeView(KBudgetView *view)
{
  pViewList->remove(view);
}

void KBudgetDoc::setURL(const KURL &url)
{
  doc_url=url;
}

void KBudgetDoc::setModified(bool status)
{
  modified = status;
  emit saveableStatus(status);
}

const KURL& KBudgetDoc::URL() const
{
  return doc_url;
}

void KBudgetDoc::slotUpdateAllViews(KBudgetView *sender)
{
  KBudgetView *w;
  if(pViewList)
  {
    for(w=pViewList->first(); w!=0; w=pViewList->next())
    {
      if(w!=sender)
        w->repaint();
    }
  }

}

bool KBudgetDoc::saveModified()
{
  bool completed=true;

  if(modified)
  {
    KBudgetApp *win=(KBudgetApp *) parent();
    int want_save = KMessageBox::warningYesNoCancel(win,
                                         i18n("The current file has been modified.\n"
                                              "Do you want to save it?"),
                                         i18n("Warning"));
    switch(want_save)
    {
      case KMessageBox::Yes:
           if (doc_url.fileName() == i18n("Untitled") || !doc_url.isValid())
           {
             win->slotFileSaveAs();
           }
           else
           {
             saveDocument(URL());
           };

           deleteContents();
           completed=true;
           break;

      case KMessageBox::No:
           modified=false;
           emit saveableStatus(true);
           deleteContents();
           completed=true;
           break;

      case KMessageBox::Cancel:
           completed=false;
           break;

      default:
           completed=false;
           break;
    }
  }

  return completed;
}

void KBudgetDoc::closeDocument()
{
  deleteContents();
}

bool KBudgetDoc::newDocument()
{
  deleteContents();
  emit documentChanged();
  doc_url.setFileName(i18n("Untitled"));
  setModified(false);

  return true;
}

bool KBudgetDoc::openDocument(const KURL& url, const char *format /*=0*/)
{
  // try to read in the file
  blockSignals(true);
  
  QString tmpfile;
  if ( !url.isLocalFile() )
    KIO::NetAccess::download(url, tmpfile);
//    KIO::NetAccess::download(url, tmpfile, (KMainWindow*)parent());
  else
    tmpfile = url.path();

  inUndo = true;
  QString error = readFile(tmpfile);
  inUndo = false;

  if ( !url.isLocalFile() )
    KIO::NetAccess::removeTempFile(tmpfile);

  blockSignals(false);
  
  if ( !error.isNull() ) {
    KMessageBox mb;
    mb.error(0, i18n("Error: %1; couldn't open %2").arg(error) \
      .arg(url.prettyURL()), i18n("Couldn't open file"));    
    return false;
  }

  // link transactions to accounts
  blockSignals(true);
  inUndo = true;
  QMap<int,Transaction>::iterator txi = transactions.begin();
  for ( ; txi != transactions.end(); txi++ ) {
    if ( isValidAccount((*txi).from()) )
      getAccount((*txi).from()).addTransaction((*txi).id());
    else
      (*txi).setFrom(-1);
    if ( isValidAccount((*txi).to()) )
      getAccount((*txi).to()).addTransaction((*txi).id());
    else
      (*txi).setTo(-1);
    }
  inUndo = false;
  blockSignals(false);

  // find next transaction and account IDs
  nextTransaction = 0;
  for ( txi = transactions.begin(); txi != transactions.end(); txi++ )
    if ( txi.key() >= nextTransaction )
      nextTransaction = txi.key() + 1;
  nextAccount = 0;
  for ( int i = 0; i < 3; i++ )
    for ( QMap<int,Account>::const_iterator acct = accounts[i].begin(); acct != accounts[i].end(); acct++ )
      if ( acct.key() >= nextAccount )
        nextAccount = acct.key() + 1;

  // gather information for recurrences
  for ( txi = transactions.begin(); txi != transactions.end(); txi++ )
    if ( (*txi).doesRecur() )
      recurrences.insert(pair<QDateTime,int>((*txi).nextNotify(), (*txi).id()));
  if ( !recurrences.empty() ) {
    int diff = QDateTime::currentDateTime().secsTo(recurrences.begin()->first);
    rTimer->start(1000*diff, true);
  }

  // do notifications and update status
  emit documentChanged();

  if ( !accounts[0].empty() || !accounts[1].empty() || !accounts[2].empty() )
    emit accountsAvailable(true);

  modified=false;
  doc_url = url;
  emit saveableStatus(false);

  return true;
}

QString KBudgetDoc::readFile(QString filename)
{
  QFile ifile(filename);
  if ( !ifile.open(IO_ReadOnly) ) {
    return i18n("Error: couldn't open %1").arg(filename);
  } else if ( !ifile.size() ) {
    // TODO: make an empty document
    return QString::null;
  }

  QTextStream inf(&ifile);
  inf.setEncoding( QTextStream::UnicodeUTF8 );
  QString type, name;
  QDomDocument doc;

  // try to detect type and validate the document
  bool flatFile = false;
  inf >> type;
  name = inf.readLine().stripWhiteSpace();
  if ( type == "KBudget" || (name == "doc 0.1" || name == "doc 0.2") )
    flatFile = true;
  else {
    // return to the start of the file before checking if it's XML, just in case
    ifile.at(0);
    QString errorMsg;
    int errorLine = 0, errorCol;
    doc.setContent(&ifile, &errorMsg, &errorLine, &errorCol);
    if ( errorLine ) {
      return i18n("not a valid KBudget file (XML parse error: \"%2\" at line %3, col %4)") \
      .arg(errorMsg).arg(errorLine).arg(errorCol);
    }
  }

  // TODO: try to validate and read in the file to temporary data structures all
  // at once; if it works, then clear the document and load the new data
  closeDocument();

  // read in document
  if ( flatFile )
    readFlatFile(inf, name);
  else {
    // check toplevel element and file version; when compatibility is broken,
    // the major version number will be changed -- this means that the file
    // versions will always be 0.x if this never happens, but that's ok.
    QDomElement root = doc.documentElement();
    if ( root.tagName() != "KBudgetDoc" )
      return i18n("invalid root element %1 in XML document").arg(root.tagName());
    QDomAttr v = root.attributeNode("version");
    
    bool ok = true;
    float version = v.value().toFloat(&ok);
    if ( v.isNull() )
      return i18n("root element has no version attribute");
    if ( !ok )
      return i18n("root element has invalid version attribute");
    if ( version >= 1.0f )
      return i18n("this file must be opened with a more recent version of KBudget");

    // now we can finally process the document
    unsigned int id, type;
    for ( QDomNode n = root.firstChild(); !n.isNull(); n = n.nextSibling() ) {
      if ( n.isElement() ) {
        QDomElement e = n.toElement();
        if ( e.tagName() == "transaction" ) {
          // TODO: print some debugging output if ID is invalid; see account too
          id = e.attributeNode("id").value().toUInt();

          // TODO: merge code for this with readFlatFile?
          if ( transactions.find(id) != transactions.end() )
            kdWarning() << "Duplicate transaction " << id << endl;
          else try {
            transactions[id] = Transaction(this, id, e);
          } catch ( std::runtime_error &e ) {
            kdWarning() << "Error reading transaction " << id << ": " << e.what() << endl;
          }
        } else if ( e.tagName() == "account" ) {
          id = e.attributeNode("id").value().toUInt();
          type = e.attributeNode("type").value().toUInt();

          if ( haveAccount(id) )
            kdWarning() << "Duplicate account " << id << endl;
          else if ( type > 2 )
            kdWarning() << "Invalid account type " << type << endl;
          else try {
            accounts[type][id] = Account(this, id, (Account::types)type, e);
          } catch ( std::runtime_error &e ) {
            kdWarning() << "Error reading account " << id << ": " << e.what() << endl;
          }
        }
      }
    }
  }

  return QString::null;
}

void KBudgetDoc::readFlatFile(QTextStream &inf, QString name)
{
  QString type;
  int version = (name == "doc 0.1" ? 1 : 2), id;
  do {
    inf >> type;
    if ( type == "Transaction" ) {
      inf >> id;
      if ( transactions.find(id) != transactions.end() ) {
        kdWarning() << "Duplicate transaction " << id << endl;
        inf.readLine();
      } else {
        transactions[id] = Transaction(this, id, inf, version);
      }
    } else if ( type == "Account" ) {
      int type;
      inf >> id >> type;
      if ( haveAccount(id) ) {
        kdWarning() << "Duplicate account " << id << endl;
        inf.readLine();
      } else {
        accounts[type][id] = Account(this, id, (Account::types)type, inf, version);
      }
    }
  } while ( !inf.atEnd() );
};

bool KBudgetDoc::saveDocument(const KURL& url, const char *format /*=0*/)
{
  if ( !url.isValid() ) {
    KMessageBox::sorry(0, i18n("%1 is a malformed URL.").arg(url.prettyURL()));
    return false;
  }

  if ( url.isLocalFile() ) {
    if ( url != doc_url && QFile::exists(url.path()) ) {
      if ( KMessageBox::warningYesNo(0, i18n("That file already exists. Would you like to overwrite the old copy?"))\
        != KMessageBox::Yes)
        return false;
    }
    return writeDocument(url.path());
  }

  KTempFile tf;
  tf.setAutoDelete(true);
  return writeDocument(tf.name());
  // TODO: upload?
}

void KBudgetDoc::deleteContents()
{
  for ( int i = 0; i < 3; i++ )
    accounts[i].clear();
  transactions.clear();
  nextAccount = 0;
  nextTransaction = 0;

  undos.clear();
  recurrences.clear();
  lastSavedPos = undos.at();
  emit undoStatus(false, false);
  emit accountsAvailable(false);
}

bool KBudgetDoc::haveAccount(int id)
{
  for ( int i = 0; i < 3; i++ )
    if ( accounts[i].find(id) != accounts[i].end() )
      return true;
  return false;
}

Account& KBudgetDoc::getAccount(int id)
{
  for ( int i = 0; i < 3; i++ )
    if ( accounts[i].find(id) != accounts[i].end() )
      return accounts[i][id];
  return blankAcct;
}

Account KBudgetDoc::getAccount(int id, QDate month)
{
  return getAccount(id).clone(month);
}

QValueVector<AccountAccessor> KBudgetDoc::getAccounts(Account::types type, const KPeriod &period)
{
  QValueVector<AccountAccessor> retList;
  QMap<int,Account>::iterator account = accounts[type].begin();
  for ( ; account != accounts[type].end(); account++ ) {
    retList.push_back(AccountAccessor(&*account, period));
  }
  return retList;
}

QPtrList<Account> KBudgetDoc::getAllAccounts(Account::types type)
{
  QPtrList<Account> ret;
/*#if QT_VERSION >= 0x030005
  // this is much cleaner code, so I'll keep it around instead of just replacing it...
  // when it's safe to drop support for early QT3 versions, this will be the only code here
  ret = accounts[0].values();
  ret += accounts[1].values();
  ret += accounts[2].values();
#else*/
  for ( int i = 0; i < 3; i++ ) {
    QMap<int,Account>::iterator ai = accounts[i].begin();
    for ( ; ai != accounts[i].end(); ai++ )
      if ( type == Account::AC_ALL || (*ai).type() == type )
        ret.append(&*ai);
  }
//#endif
  return ret;
}

Transaction& KBudgetDoc::getTransaction(const int id)
{
  if ( transactions.find(id) != transactions.end() )
    return transactions[id];
  return blankTransaction;
}

QValueVector<Transaction*> KBudgetDoc::getTransactions(const KPeriod &period)
{
  QValueVector<Transaction*> list;
  for ( QMap<int,Transaction>::iterator it = transactions.begin(); it != transactions.end(); it++ ) {
    if ( !period.contains((*it).date()) )
      continue;
    Transaction& tx = *it;
    QValueVector<Transaction*>::iterator next = list.end();
    // while there is a previous item and the previous item should be after the new
    // item, make the next item equal to the previous item
    while ( next != list.begin() && (*(next-1))->date() > tx.date() )
      next--;
    // insert before the next item
    list.insert(next, &tx);
  }
  return list;
}

int KBudgetDoc::newAccount(Account::types type, QString name, float initialBalance)
{
  // NOTE: the account id must always increase; if an account is deleted, the
  // transactions that belong to it will still point to its id, and a new
  // account using this id could accidentally acquire extra transactions. The
  // transactions must keep the old id to make undo/redo easier.

  accounts[type].insert(nextAccount, Account(this, nextAccount, type, name, initialBalance));
  nextAccount++;

  if ( !inUndo )
    addUndo(new KBAccountUndo(this, KBAccountUndo::UA_ADD, nextAccount - 1));

  emit accountsAvailable(true);    
  emit accountChanged(nextAccount - 1, Added);
  modified=true;
  emit saveableStatus(true);
  return nextAccount - 1;
}

void KBudgetDoc::deleteAccount(int id)
{
  if ( id < 0 )
    return;

  QMap<int,Account>::iterator acct;
  int i;
  for ( i = 0; i < 3; i++ ) {
    acct = accounts[i].find(id);
    if ( acct != accounts[i].end() )
      break;
  }
  if ( acct == accounts[2].end() )
    return;

  if ( !inUndo )
    addUndo(new KBAccountUndo(this, KBAccountUndo::UA_DEL, id));

  emit accountChanged(id, Deleted);
  accounts[i].remove(acct);
  emit accountDeleted(id);
  if ( accounts[0].empty() && accounts[1].empty() && accounts[2].empty() )
    emit accountsAvailable(false);
  setModified(true);
}

bool KBudgetDoc::isValidAccount(int id)
{
  for ( int i = 0; i < 3; i++ )
    if ( accounts[i].find(id) != accounts[i].end() )
      return true;
  return false;
}

int KBudgetDoc::newTransaction(int from, int to, float value, QDate date, QString description)
{
  transactions.insert(nextTransaction, Transaction(this, nextTransaction, from, to, value, date, description));
  getAccount(from).addTransaction(nextTransaction);
  getAccount(to).addTransaction(nextTransaction);

  if ( !inUndo )
    addUndo(new KBTransactionUndo(this, KBTransactionUndo::UT_ADD, nextTransaction));

  nextTransaction++;
  emit accountChanged(from, Modified);
  emit accountChanged(to, Modified);
  emit transactionChanged(nextTransaction - 1, Added);
  modified=true;
  emit saveableStatus(true);
  return nextTransaction - 1;
}

void KBudgetDoc::deleteTransaction(int id)
{
  if ( id < 0 )
    return;
    
  if ( transactions.find(id) != transactions.end() ) {
    if ( !inUndo )
      addUndo(new KBTransactionUndo(this, KBTransactionUndo::UT_DEL, id));

    Transaction &tx = transactions[id];
    if ( haveAccount(tx.from()) )
      getAccount(tx.from()).removeTransaction(id);
    if ( haveAccount(tx.to()) )
      getAccount(tx.to()).removeTransaction(id);

    emit transactionChanged(id, Deleted);

    transactions.remove(id);
    for ( multimap<QDateTime,int>::iterator it = recurrences.begin(); it != recurrences.end(); it++ )
      if ( it->second == id ) {
        recurrences.erase(it);
        break;
      }
    emit transactionDeleted(id);
    modified=true;
    emit saveableStatus(true);
  }
}

void KBudgetDoc::undo()
{
  inUndo = true;
  if ( undos.current() ) {
    undos.current()->undo();
    emit undoStatus( undos.prev() ? true : false, true);
  }
  if ( (lastSavedPos < 0 && !undos.current()) || lastSavedPos == undos.at() )
    setModified(false);
  else
    setModified(true);
  inUndo = false;
}

void KBudgetDoc::redo()
{
  inUndo = true;
  if ( undos.current() ) {
    undos.next()->redo();
  } else if ( undos.first() ) {
    undos.first()->redo();
  } else {
    emit undoStatus(false, false);
    return;
  }

  emit undoStatus( true, undos.current() != undos.getLast() ? true : false);
  if ( (lastSavedPos < 0 && !undos.current()) || lastSavedPos == undos.at() )
    setModified(false);
  else
    setModified(true);
  inUndo = false;
}

bool KBudgetDoc::canAddUndo()
{
  return !inUndo;
}

void KBudgetDoc::beginUndoGroup()
{
  if ( !ugroup )
    ugroup = new KBUndoGroup(this);
}

void KBudgetDoc::endUndoGroup()
{
  if ( !ugroup )
    return;

  KBUndoGroup* grp = ugroup;
  ugroup = 0;
  addUndo(grp);
}

int KBudgetDoc::suggestFrom(int account)
{
  if ( !haveAccount(account) )
    return -1;

  Account::types type = getAccount(account).type();
  if ( type == Account::AC_EXPENSE )
    return (*accounts[2].begin()).id();
  return account;
}

int KBudgetDoc::suggestTo(int account)
{
  if ( !haveAccount(account) )
    return -1;

  Account::types type = getAccount(account).type();
  if ( type == Account::AC_INCOME )
    return (*accounts[2].begin()).id();
  return account;
}

void KBudgetDoc::slotCheckRecurrences()
{
  // I would rather do this in a loop, but that doesn't work if the signal
  // returns before the processing is finished (which happens with the current
  // one-shot dialog box). Instead, we can take advantage of the fact that
  // the recurrence queue will be updated when the processing is done, setting
  // up the timer to call this slot again.
  kdDebug() << "There are currently " << recurrences.size() << " recurring transactions." << endl;
  while ( !(recurrences.empty() || recurrences.begin()->first.isValid()) )
    recurrences.erase(recurrences.begin());
  if ( recurrences.empty() )
    return;

  QDateTime now = QDateTime::currentDateTime();
  pair<QDateTime, int> t = *recurrences.begin();
  kdDebug() << "Checking transaction " << t.second << "; next notification: " << t.first.toString(Qt::ISODate) << "(in " << now.secsTo(t.first) << "secs)" << endl;
  // do this to adjust for differences in millisecond times
  if ( now.secsTo(t.first) <= 1 ) {
    kdDebug() << "Instantiating transaction " << t.second << endl;
    emit instantiateRecurrence(t.second);
  }
}

void KBudgetDoc::signalAcctChanged(int id)
{
  modified=true;
  emit saveableStatus(true);
  emit accountChanged(id, Modified);
}

void KBudgetDoc::signalTransactionChanged(int id)
{
  modified=true;
  emit saveableStatus(true);
  emit transactionChanged(id, Modified);
}

void KBudgetDoc::signalTransactionRecurrenceChanged(int id)
{
  modified=true;
  emit saveableStatus(true);

  // check if it's already in the list; if so, remove it
  multimap<QDateTime,int>::iterator it;
  for ( it = recurrences.begin(); it != recurrences.end(); it++ ) {
    if ( it->second == id ) {
      recurrences.erase(it);
      break;
    }
  }

  Transaction &tr = getTransaction(id);
  if ( tr.doesRecur() ) {
    // now add it at the correct date
    recurrences.insert(pair<QDateTime,int>(tr.nextNotify(), tr.id()));

    // set the timer for the next recurrence
    int diff = QDateTime::currentDateTime().secsTo(recurrences.begin()->first);
    rTimer->start(1000*diff, true);
  }
}

bool KBudgetDoc::writeDocument(QString file)
{
  QFileInfo info(file);
  if ( info.isDir() ) {
    KMessageBox::sorry(0, i18n("You selected a directory!"));
    return false;
  }

  KSaveFile ofile(file, 0600);
  if ( ofile.status() ) {
    KMessageBox::sorry(0, i18n("Couldn't open file for writing"));
    ofile.abort();
    return false;
  }

  QTextStream &outf = *ofile.textStream();
  outf.setEncoding( QTextStream::UnicodeUTF8 );
  QDomDocument doc("KBudgetDoc");
  doc.appendChild( doc.createProcessingInstruction( "xml", "version=\"1.0\" encoding=\"UTF-8\"" ) );
  QDomElement root = doc.createElement("KBudgetDoc");
  root.setAttribute("version", "0.3");
  doc.appendChild(root);

  QMap<int,Transaction>::iterator t = transactions.begin();
  for ( ; t != transactions.end(); t++ )
    // if the from and to accounts don't exist, don't write the transaction
    if ( haveAccount((*t).from()) || haveAccount((*t).to()) )
      (*t).serialize(root);

  QMap<int,Account>::iterator acct;
  for ( int i = 0; i < 3; i++ )
    for ( acct = accounts[i].begin(); acct != accounts[i].end(); acct++ )
      (*acct).serialize(root);
  
  outf << doc.toString();

  if ( ofile.status() ) {
    KMessageBox::error(0, i18n("Error while writing file; file was not saved"));
    ofile.abort();
    return false;
  }

  ofile.close();
  if ( ofile.status() ) {
    KMessageBox::sorry(0, i18n("Error after saving file; data may not have been saved"));
    return false;
  }

  lastSavedPos = undos.at();
  modified=false;
  emit saveableStatus(false);
  return true;
}

void KBudgetDoc::insertAccount(int id, Account* copy)
{
  accounts[(int)copy->type()][id] = *copy;
  emit accountsAvailable(true);
  emit accountChanged(id, Added);
}

void KBudgetDoc::insertTransaction(int id, Transaction* copy)
{
  transactions[id] = *copy;
  if ( haveAccount(copy->from()) )
    getAccount(copy->from()).addTransaction(id);
  if ( haveAccount(copy->to()) )
    getAccount(copy->to()).addTransaction(id);
  emit transactionChanged(id, Added);
  signalTransactionRecurrenceChanged(id);
}

void KBudgetDoc::addUndo(KBUndoAction* a)
{
  if ( ugroup ) {
    ugroup->add(a);
    return;
  }

  /** if the current item is not the last, remove everything after the current item */
  int remove = undos.count() - 1 - undos.at();
  if ( remove > 0 )
    for ( int i = 0; i < remove; i++ )
      undos.removeLast();

  /** append the new item */
  undos.append(a);

  /** the current item is the last one, so we can go back but not forward */
  emit undoStatus(true, false);
}
