/***************************************************************************
                          kbudgetview.cpp  -  description
                             -------------------
    begin                : Sat Aug 10 16:03:35 CST 2002
    copyright            : (C) 2002 by Richard Garand
    email                : richard@garandnet.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// system includes
#include <time.h>
#include <string.h>
#include <stdio.h>

// include files for Qt
#include <qprinter.h>
#include <qpainter.h>
#include <klocale.h>
#include <qlayout.h>
#include <qpopupmenu.h>
#include <kstatusbar.h>
#include <qheader.h>

// application specific includes
#include "kbudgetview.h"
#include "kbudgetdoc.h"
#include "kbudget.h"
#include "kbudgeteditor.h"
#include "ktransactionwindow.h"
#include "knewaccount.h"
#include "knewtransaction.h"
#include "kbreconcilewindow.h"

class KBudgetViewItem : public QListViewItem
{
  public:
    /** makes an empty row */
    KBudgetViewItem(QListView* parent);
    /** makes a row after after, with account accessor acct, and updates total row total */
    KBudgetViewItem(QListView* parent, KBudgetViewItem* after, const AccountAccessor& acct, KBudgetViewItem* total);
    /** makes a row with preset text and initializes numbers to 0 */
    KBudgetViewItem(QListView* parent, QString s1, QString s2 = QString(), QString s3 = QString(), QString s4 = QString());

    int id() const;
    /** returns the requested value; i in [0, 2] */
    float value(unsigned int i) const;

    /** returns a reference to the account accessor */
    AccountAccessor& accessor();

    /** sets the accessor to use; has no effect if a valid accessor is not already in use */
    void setAccessor(AccountAccessor& acct);
    /** refreshes the accessor and updates values */
    void reread();
    
    /** Removes the item's values from the relevant total line; do this before
      * deleting a single item. */
    void remove();
  private:
    /** gets values from the accessor and updates the totals */
    void getValues();
    /** sets the column text from numbers[] */
    void refresh();

    /** account accessor to use if this row holds an account; 0 otherwise */
    AccountAccessor acct;
    /** the numbers (if this row holds an account) */
    float numbers[3];
    /** the totals row for this section (if this row holds an account) */
    KBudgetViewItem* total;
};

KBudgetViewItem::KBudgetViewItem(QListView* parent)
  : QListViewItem(parent)
{
}

KBudgetViewItem::KBudgetViewItem(QListView* parent, KBudgetViewItem* after, const AccountAccessor& a, KBudgetViewItem* t)
  : QListViewItem(parent, after), acct(a), total(t)
{
  setText(0, acct.name());

  for ( int i = 0; i < 3; i++ )
    numbers[i] = 0;

  getValues();
}

KBudgetViewItem::KBudgetViewItem(QListView* parent, QString s1, QString s2, QString s3, QString s4)
  : QListViewItem(parent, s1, s2, s3, s4)
{
  for ( int i = 0; i < 3; i++ )
    numbers[i] = 0;
}

int KBudgetViewItem::id() const
{
  return acct.id();
}

float KBudgetViewItem::value(unsigned int i) const
{
  if ( i > 2 )
    return 0;
  return numbers[i];
}

void KBudgetViewItem::setAccessor(AccountAccessor& a)
{
  if ( acct.id() == -1 )
    return;

  acct = a;

  getValues();
}

void KBudgetViewItem::remove()
{
  if ( acct.id() != -1 ) {
    for ( int i = 0; i < 3; i++ )
      total->numbers[i] -= numbers[i];
    total->refresh();
  }
}

void KBudgetViewItem::reread()
{
  acct.refresh();
  setText(0, acct.name());
  getValues();
}

void KBudgetViewItem::getValues()
{
  // remove numbers from totals
  for ( int i = 0; i < 3; i++ )
    total->numbers[i] -= numbers[i];

  // get the data
  if ( acct.type() != Account::AC_CASH ) {
    numbers[0] = acct.startingBudget() + acct.budgeted();
    numbers[1] = acct.change();
    numbers[2] = acct.budget();
  } else {
    numbers[0] = acct.startingBalance();
    numbers[2] = acct.balance();
    numbers[1] = numbers[2] - numbers[0];
  }

  // refresh columns
  refresh();

  // update the totals
  for ( int i = 0; i < 3; i++ )
    total->numbers[i] += numbers[i];
  total->refresh();
}

void KBudgetViewItem::refresh()
{
  setText(1, KGlobal::locale()->formatMoney(numbers[0]));
  setText(2, KGlobal::locale()->formatMoney(numbers[1]));
  setText(3, KGlobal::locale()->formatMoney(numbers[2]));
}
                                
const int COLUMNS = 4;
const int ID_EXPECTED_CHANGE = 2;
const int ID_EXPECTED_BALANCE = 3;

KBudgetView::KBudgetView(QWidget *parent, const char *name) : QVBox(parent, name), selAccount(-1)
{
  setBackgroundMode(PaletteBase);

  period = new KPeriodSelector(this);

  lv = new QListView(this, "kbudgetview_listview");
  for ( int i = 0; i < COLUMNS; i++ ) {
    lv->addColumn("");
    if ( i )
      lv->setColumnAlignment(i, Qt::AlignRight);
  }
  setStretchFactor(lv, 1);
  lv->setSorting(-1);
  lv->header()->setStretchEnabled(true);
  lv->setAllColumnsShowFocus(true);

  QString st = KGlobal::locale()->formatMoney(0);

  // create initial layout  
  totals[2] = new KBudgetViewItem(lv, i18n("Total"), st, st, st);
  headers[2] = new KBudgetViewItem(lv, i18n("Account"), i18n("Start"), i18n("account balance change", "Change"), i18n("End"));
  new KBudgetViewItem(lv, i18n("Accounts"));
  new KBudgetViewItem(lv);
  totals[1] = new KBudgetViewItem(lv, i18n("Total"), st, st, st);
  headers[1] = new KBudgetViewItem(lv, i18n("Account"), i18n("Budgeted"), i18n("Spent"), i18n("Remaining"));
  new KBudgetViewItem(lv, i18n("Expenses"));
  new KBudgetViewItem(lv);
  totals[0] = new KBudgetViewItem(lv, i18n("Total"), st, st, st);
  headers[0] = new KBudgetViewItem(lv, i18n("Account"), i18n("Budgeted"), i18n("Received"), i18n("Remaining"));
  new KBudgetViewItem(lv, i18n("Income"));

  editor = new KBudgetEditor(this, getDocument());

  connect(getDocument(), SIGNAL(accountChanged(int,KBudgetDoc::ChangeAction)), SLOT(slotAccountChanged(int,KBudgetDoc::ChangeAction)));
  connect(getDocument(), SIGNAL(documentChanged()), SLOT(slotDocumentChanged()));
  connect(period, SIGNAL(periodChanged(KPeriod)), SLOT(slotPeriodChanged(KPeriod)));
  connect(lv, SIGNAL(selectionChanged(QListViewItem*)), SLOT(slotSelectionChanged(QListViewItem*)));
  connect(lv, SIGNAL(contextMenuRequested(QListViewItem*, const QPoint&, int)), SLOT(slotContextMenuRequested(QListViewItem*, const QPoint&)));
  // TODO: use KListView::executed()
  connect(lv, SIGNAL(doubleClicked(QListViewItem*)), SLOT(slotTransactions()));
  connect(lv, SIGNAL(currentChanged(QListViewItem*)), SLOT(slotSetSelected(QListViewItem*)));

  QDate today = QDate::currentDate();
  selectMonth(today.year(), today.month());

  KStatusBar *sb = ((KMainWindow*)parent)->statusBar();
  sb->insertItem(i18n("Expected change: %1").arg(KGlobal::locale()->formatMoney(0)), ID_EXPECTED_CHANGE);
  sb->insertItem(i18n("Expected balance: %1").arg(KGlobal::locale()->formatMoney(0)), ID_EXPECTED_BALANCE);
}

KBudgetView::~KBudgetView()
{
}

void KBudgetView::selectMonth(int yr, int mn)
{
  // the signal activated by this takes care of everything
  period->setPeriod(KPeriod(yr, mn));
}

KBudgetDoc *KBudgetView::getDocument() const
{
  KBudgetApp *theApp=(KBudgetApp *) parentWidget();

  return theApp->getDocument();
}

void KBudgetView::print(QPrinter *pPrinter)
{
  QPainter printpainter;
  printpainter.begin(pPrinter);

  // TODO: add your printing code here

  printpainter.end();
}

int KBudgetView::currentAccount() const
{
  return selAccount;
}

void KBudgetView::slotPeriodChanged(KPeriod to)
{
  // get accounts
  AccessorVector accounts[3];
  for ( int i = 0; i < 3; i++ )
    accounts[i] = getDocument()->getAccounts((Account::types)i, period->period());

  // update items
  for ( int i = 0; i < 3; i++ )
    for ( unsigned int j = 0; j < accounts[i].size(); j++ )
      if ( items.find(accounts[i][j].id()) != items.end() )
        items[accounts[i][j].id()]->setAccessor(accounts[i][j]);

  // update the status bar
  updateStatusBar();

  // change the editor to the new period
  editor->selectPeriod(to);
}

void KBudgetView::slotDocumentChanged()
{
  // clear summary
  for ( QMap<int, KBudgetViewItem*>::iterator it = items.begin(); it != items.end(); it++ ) {
    (*it)->remove();
    delete *it;
  }
  items.clear();

  // for each section
  AccessorVector accounts[3];
  for ( int i = 0; i < 3; i++ ) {
    // get accounts
    accounts[i] = getDocument()->getAccounts((Account::types)i, period->period());
    // clear itemlist
    itemlist[i].clear();
  }

  createSummary(accounts);
  updateStatusBar();
  editor->hide();
}

void KBudgetView::slotAccountChanged(int acct, KBudgetDoc::ChangeAction change)
{
  Account& account = getDocument()->getAccount(acct);
  int type = (int)account.type();

  switch ( change ) {
  case KBudgetDoc::Added:
    if ( itemlist[type].empty() )
      // add it at the beginning
      addAccount(type, 0, account, headers[type]);
    else if ( itemlist[type].back()->id() < acct )
      // add it at the end
      addAccount(type, itemlist[type].size(), account, itemlist[type].back());
    else
      // insert it before the first account with a greater id
      for ( unsigned int i = 0; i < itemlist[type].size(); i++ )
        if ( itemlist[type][i]->id() > acct ) {
          addAccount(type, i, account, i ? itemlist[type][i-1] : headers[type]);
          break;
        }
    break;
    
  case KBudgetDoc::Deleted:
    // remove
    for ( ViewItemVector::iterator i = itemlist[type].begin(); i != itemlist[type].end(); i++ )
      if ( (*i)->id() == acct ) {
        itemlist[type].erase(i);
        break;
      }

    if ( acct == selAccount ) {
      ((KBudgetViewItem*)lv->selectedItem())->remove();
      delete lv->selectedItem();
    } else
      for ( QListViewItemIterator it(lv); it.current(); it++ )
        if ( ((KBudgetViewItem*)it.current())->id() == acct ) {
          ((KBudgetViewItem*)it.current())->remove();
          delete it.current();
          break;
        }
    break;
    
  case KBudgetDoc::Modified:
    if ( items.find(acct) != items.end() )
      items[acct]->reread();
    break;
  }

  updateStatusBar();
}

void KBudgetView::slotSelectionChanged(QListViewItem* item)
{
  if ( !item ) {
    selAccount = -1;
    return;
  }

  selAccount = ((KBudgetViewItem*)item)->id();
  if ( selAccount == -1 ) {
    editor->hide();
    emit accountSelectionStatus(false);
  } else {
    if ( getDocument()->getAccount(selAccount).type() != Account::AC_CASH ) {
      editor->selectAccount(selAccount);
      editor->show();
    } else
      editor->hide();
    emit accountSelectionStatus(true);
  }
}

void KBudgetView::slotContextMenuRequested(QListViewItem* item, const QPoint& pos)
{
  emit contextMenuRequested(pos);
  
  if ( item );
}

void KBudgetView::slotTransactions()
{
  if ( selAccount >= 0 ) {
    KTransactionWindow* w = new KTransactionWindow(getDocument(), period->period(), selAccount);
    w->show();
  }
}

void KBudgetView::slotAllTransactions()
{
  KTransactionWindow* w = new KTransactionWindow(getDocument(), period->period(), -1);
  w->show();
}

void KBudgetView::slotEditAccount()
{
  if ( selAccount >= 0 )
    (new KNewAccount((KMainWindow*)parent(), getDocument(), selAccount))->show();
}

void KBudgetView::slotNewAccount()
{
  if ( selAccount >= 0 )
    (new KNewAccount((KMainWindow*)parent(), getDocument(), -1, getDocument()->getAccount(selAccount).type()))->show();
  else
    (new KNewAccount((KMainWindow*)parent(), getDocument(), -1, Account::AC_INCOME))->show();
}

void KBudgetView::slotNewTransaction()
{
  (new KNewTransaction((KMainWindow*)parent(), getDocument(), selAccount))->show();
}

void KBudgetView::slotReconcileAccount()
{
  if ( selAccount >= 0 )
    (new KBReconcileWindow(getDocument(), selAccount))->show();
}

void KBudgetView::slotSetSelected(QListViewItem* i)
{
  lv->setSelected(i, true);
}

void KBudgetView::createSummary(AccessorVector accounts[3])
{
  // add accounts
  for ( int section = 0; section < 3; section++ ) {
    KBudgetViewItem* after = headers[section];
    for ( unsigned int i = 0; i < accounts[section].size(); i++ ) {
      after = new KBudgetViewItem(lv, after, accounts[section][i], totals[section]);
      items[after->id()] = after;
      itemlist[section].push_back(after);
    }
  }

  // handle selection
  slotSelectionChanged(lv->selectedItem());
}

void KBudgetView::updateStatusBar()
{
  float sums[3] = {totals[0]->value(0), totals[1]->value(0), totals[2]->value(0)};

  KStatusBar *sb = ((KMainWindow*)parent())->statusBar();
  sb->changeItem(i18n("Expected change: %1").arg(KGlobal::locale()->formatMoney(sums[0] - sums[1])), ID_EXPECTED_CHANGE);
  sb->changeItem(i18n("Expected balance: %1").arg(KGlobal::locale()->formatMoney(sums[2] + sums[0] - sums[1])), ID_EXPECTED_BALANCE);
}

void KBudgetView::addAccount(int type, int pos, Account& account, KBudgetViewItem* previous)
{
  // create item and put it in itemlist
  KBudgetViewItem* item = new KBudgetViewItem(lv, previous, AccountAccessor(&account, period->period()), totals[type]);
  itemlist[type].insert(itemlist[type].begin()+pos, item);
  items[account.id()] = item;
}
