/***************************************************************************
                          kperiod.h  -  description
                             -------------------
    begin                : Sun Dec 22 2002
    copyright            : (C) 2002 by Richard Garand
    email                : richard@garandnet.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KPERIOD_H
#define KPERIOD_H

class QDate;

/** The KPeriod class represents a range of dates. There are currently three
  * types of periods: single months, date ranges, and "all". Some functions only
  * work for "unit" periods - at the moment, this means the month type; the
  * "all" period also makes some functions meaningless.
  *
  * When a period is created with the default constructor, it gest the "all"
  * type; other constructors can be used to directly create another type of
  * period.
  *
  * All inidividual values used should be the same as QDate would use;
  * months are 1-12, days are 1-31, and years must be valid.
  * @author Richard Garand
  */

class KPeriod {
  public:
    enum Types {All, Month, DateRange};
  
    KPeriod();
    /** Creates a period with a given month selected */
    KPeriod(int year, int month);
    /** Creates a period containing a range of dates (including from and to). */
    KPeriod(QDate from, QDate to);
    ~KPeriod();
    
    /** select a given month (month number is 1-based) */
    void selectMonth(int year, int month);
    /** change the year */
    void addYears(int years);
    /** change the month */
    void addMonths(int months);

    /** select the previous period (only works for unit periods) */
    void previous();
    /** select the next period (only works for unit periods) */
    void next();

    /** return the first year */
    int firstYear() const;
    /** return the number of the first month */
    int firstMonth() const;
    /** returns the first day of the period */
    QDate firstDay() const;
    
    /** Returns the length of the period in days. */
    int length() const;
    
    /** return the name or description of the period */
    QString name() const;

    /** returns true if the given date is in the period */
    bool contains(const QDate date) const;
    /** returns true if the period contains any days in the given year */
    bool containsYear(const int year) const;
    /** returns true if the period contains any days in the given month */
    bool containsMonth(const int year, const int month) const;

    /** returns the type of period */
    Types type() const;
  private:
    QDate m_from, m_to;
    Types m_type;
};

#endif
