/***************************************************************************
                          ktransactionlist.cpp  -  description
                             -------------------
    begin                : Sat Sep 7 2002
    copyright            : (C) 2002 by Richard Garand
    email                : richard@garandnet.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qlabel.h>
#include <qdatetime.h>

#include <klistview.h>
#include <klocale.h>
#include <kdebug.h>
 
#include "ktransactionlist.h"
#include "kbudgetdoc.h"
#include "ktransactionlistitem.h"
#include "knewtransaction.h"
#include "kperiodselector.h"

KTransactionList::KTransactionList(KBudgetDoc* doc, QWidget *parent, int account, Profiles profile)
  : KListView(parent,"transaction_list"), m_doc(doc), m_account(account), m_initialBalance(0),
  m_allAccounts(false), m_showBalance(true), total(0), first(0)
{
  Account& acct = m_doc->getAccount(m_account);

  if ( account == -1 ) {
    m_allAccounts = true;
    m_showBalance = false;
  }
  if ( profile == NormalList && acct.type() != Account::AC_CASH )
    m_showBalance = false;

  // initialize column index array
  int add = 0;
  for ( int i = 0; i <= 2; i++ )
    colIndices[i] = i + add;
  if ( m_allAccounts ) add-=2;
  else add--;
  colIndices[3] = 3 + add;
  if ( !m_showBalance ) add--;
  colIndices[4] = 4 + add;
  if ( !m_allAccounts ) add -= 2;
  for ( int i = 5; i <= 7; i++ )
    colIndices[i] = i + add;

  if ( !m_showBalance ) {
    colIndices[4] = -1;
  }

  // add columns to list view
  addColumn(i18n("Date"));
  if ( !m_allAccounts ) {
    addColumn(i18n("Debits"));
    setColumnAlignment(1, Qt::AlignRight);
    addColumn(i18n("Credits"));
    setColumnAlignment(2, Qt::AlignRight);
  } else {
    addColumn(i18n("Value"));
    setColumnAlignment(1, Qt::AlignRight);
  }
  if ( m_showBalance ) {
    addColumn(i18n("Balance"));
    setColumnAlignment(colIndices[(int)Balance], Qt::AlignRight);
  }
  if ( m_allAccounts ) {
    addColumn(i18n("From"));
    addColumn(i18n("To"));
  }
  addColumn(i18n("Description"));

  setShowSortIndicator(true);
  setAllColumnsShowFocus(true);

  if ( profile == Reconcile )
    setSelectionMode(QListView::Multi);

  connect(this, SIGNAL(contextMenuRequested(QListViewItem*,const QPoint&,int)), SLOT(slotContextMenuRequested(QListViewItem*, const QPoint&, int)));
  connect(this, SIGNAL(selectionChanged()), SLOT(slotSelectionChanged()));
  connect(this, SIGNAL(currentChanged(QListViewItem*)), SLOT(slotCurrentChanged(QListViewItem*)));
  connect(this, SIGNAL(doubleClicked(QListViewItem*, const QPoint&, int)), SLOT(slotEditTransaction()));
  connect(m_doc, SIGNAL(transactionChanged(int,KBudgetDoc::ChangeAction)), SLOT(slotTransactionChanged(int,KBudgetDoc::ChangeAction)));
}

KTransactionList::~KTransactionList()
{
}

bool KTransactionList::showBalance() const
{
  return m_showBalance;
}

bool KTransactionList::showAllAccounts() const
{
  return m_allAccounts;
}

int KTransactionList::getColumnIndex(Columns col) const
{
  return colIndices[(int)col];
}

int KTransactionList::getCurrent()
{
  if ( (KTransactionListItem*)selectedItem() )
    return ((KTransactionListItem*)selectedItem())->tid();
  else
    return -1;
}

KBudgetDoc* KTransactionList::getDocument()
{
  return m_doc;
}

void KTransactionList::clear()
{
  KListView::clear();
  items.clear();

  total = 0;
}

void KTransactionList::addTransaction(int id)
{
  items[id] = new KTransactionListItem(this, &m_doc->getTransaction(id), 0, m_account, total);
}

void KTransactionList::slotPeriodChanged(KPeriod period)
{
  clear();

  // make separator row and total row
  if ( !m_allAccounts ) {
    new KTransactionListItem(this);
    total = new KTransactionListItem(this, 0, 0);
  }
  
  KTransactionListItem* balanceItem=0;
  Account& acct = m_doc->getAccount(m_account);
  float balance = AccountAccessor(&acct, period).startingBalance();

  first = 0;
  m_initialBalance = balance;
  if ( m_showBalance )
    balanceItem = new KTransactionListItem(this, period.firstDay(), m_initialBalance);

  // create the transaction list
  QValueVector<Transaction*> txns;
  if ( m_account >= 0 ) {
    QValueList<int> txids;
    txids = acct.getTransactions(period);
    for ( unsigned int i = 0; i < txids.size(); i++ )
      txns.push_back(&m_doc->getTransaction(txids[i]));
  } else
    txns = m_doc->getTransactions(period);

  float dTotal=0, cTotal=0;
  for ( unsigned int i = 0; i < txns.size(); i++ ) {
    Transaction& tx = *txns[i];

    // update totals
    if ( tx.from() == m_account ) {
      dTotal += tx.value();
      balance -= tx.value();
    } else if ( tx.to() == m_account ) {
      cTotal += tx.value();
      balance += tx.value();
    }

    // create the new item
    items[tx.id()] = new KTransactionListItem(this, &tx, balance, m_account, total);

    // set the previous item's next
    if ( i )
      items[txns[i-1]->id()]->setNext(items[txns[i]->id()]);

    // set first if necessary
    if ( !first )
      first = items[txns[i]->id()];
  }

  // create the initial balance row if necessary
  if ( m_showBalance ) {
    balanceItem->setNext(first);
    first = balanceItem;
  }

  setSorting(getColumnIndex(Date));
  sort();

  slotSelectionChanged();

  m_period = period;
}

void KTransactionList::slotDeleteTransaction()
{
  KTransactionListItem* i = (KTransactionListItem*)currentItem();
  if ( i )
    m_doc->deleteTransaction(i->tid());
}

void KTransactionList::slotTransactionChanged(int id, KBudgetDoc::ChangeAction action)
{
  Transaction &tr = m_doc->getTransaction(id);
  ListItemMap::iterator it = items.find(id);

  if ( !m_period.contains(tr.date()) ) {
    // if it's before the selected period, update balances
    if ( tr.date() < m_period.firstDay() && first ) {
      Account& acct = m_doc->getAccount(m_account);
      float balance = AccountAccessor(&acct, m_period).startingBalance();

      float test = balance - m_initialBalance;
      kdDebug() << "KTransactionList: adding " << test << " to balance" << endl;
      first->add(balance - m_initialBalance);
      m_initialBalance = balance;
    }
    // remove it if we have it in our list
    if ( it != items.end() ) {
        delete *it;
        items.remove(it);    
    }
    return;
  }

  float debits=0, credits=0;
  if ( tr.from() == m_account )
    debits = tr.value();
  else if ( tr.to() == m_account )
    credits = tr.value();

  KTransactionListItem* item;
  KTransactionListItem* prev = first;
  switch ( action ) {
  case KBudgetDoc::Added:
    if ( m_account != -1 && tr.from() != m_account && tr.to() != m_account )
      break;
    item = new KTransactionListItem(this, &tr, 0, m_account, total);
    items[id] = item;

    // if there are other items, add the new one to the linked list
    // try to find largest one smaller than item and use insertAfter
    if ( prev ) {
      // while next is smaller...
      while ( prev->next() && prev->next()->compare(item, 0, true) < 0 )
        prev = prev->next();
      // make sure that prev really is smaller
      if ( prev->compare(item, 0, true) < 0 )
        item->insertAfter(prev);
      // there were no larger items, so item is the smallest...
      else {
        item->setNext(first);
        first = item;
      }
    }
    item->reread();
    break;
  case KBudgetDoc::Deleted:
    if ( it != items.end() ) {
      delete *it;
      items.remove(it);
    }
    break;
  case KBudgetDoc::Modified:
    // this signal can be emitted when a transactions accounts have changed,
    // so we may have to remove it or add it. This is just a quick hack to
    // fix up changes to from/to; TODO: make it work for everything; date
    // changes are another problem
    if ( it != items.end() ) {
      if ( m_account < 0 || tr.from() == m_account || tr.to() == m_account )
        (*it)->reread();
      else {
        delete *it;
        items.remove(it);
      }
    } else {
      // not in the list yet... but it should be
      slotTransactionChanged(id, KBudgetDoc::Added);
    }
    break;
  }

  // force a balanceChanged() signal
  slotSelectionChanged();
}

void KTransactionList::slotContextMenuRequested(QListViewItem*, const QPoint &p, int)
{
  emit contextMenuRequested(p);
}

void KTransactionList::slotSelectionChanged()
{
  if ( selectionMode() == QListView::Single ) {
    if ( selectedItem() )
      emit transactionFuncsEnabled(true);
    else
      emit transactionFuncsEnabled(false);
  } else {
    // get the total balance
     float totalBalance = m_initialBalance;

    for ( KTransactionListItem* item = first; item; item = item->next() ) {
      if ( !isSelected(item) )
        totalBalance += item->balanceChange();
    }

    emit balanceChanged(totalBalance);
  }
}

void KTransactionList::slotCurrentChanged(QListViewItem* i)
{
  if ( selectionMode() != QListView::Single || !i )
    return;

  // if i is not a transaction item but the item below it is...
  if ( ((KTransactionListItem*)i)->tid() == -1 && i->itemBelow() \
        && ((KTransactionListItem*)i->itemBelow())->tid() != -1 )
    setSelected(i->itemBelow(), true);
  else
    // i is a transaction item
    setSelected(i, true);
}

void KTransactionList::slotEditTransaction()
{
  KTransactionListItem* selected = (KTransactionListItem*)currentItem();
  if ( selected && selected->tid() >= 0 )
    (new KNewTransaction((KMainWindow*)parent(), m_doc, -1, selected->tid()))->show();
}

#include "ktransactionlist.moc"
