/***************************************************************************
                          transaction.cpp  -  description
                             -------------------
    begin                : Sat Aug 10 2002
    copyright            : (C) 2002 by Richard Garand
    email                : richard@garandnet.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdexcept>

#include <qtextstream.h>

#include <kdebug.h>

#include "transaction.h"
#include "kbudgetdoc.h"
#include "kbundoaction.h"

Transaction::Transaction(KBudgetDoc* doc)
  : m_doc(doc), m_from(-1), m_to(-1)
{
}

Transaction::Transaction(KBudgetDoc* doc, int id, QTextStream &stream, int version)
  : m_doc(doc), m_id(id), m_from(-1), m_to(-1)
{
  int year, month, day;
  stream >> m_from >> m_to >> m_value >> year >> month >> day;
  m_date.setYMD(year, month, day);
  m_description = stream.readLine().stripWhiteSpace();

  m_recurrence.setRecurStart(m_date);
  
  if ( version == 2 );
}

Transaction::Transaction(KBudgetDoc* doc, int id, QDomElement element)
  : m_doc(doc), m_id(id), m_from(-1), m_to(-1)
{
  bool ok;

  m_from = element.attributeNode("from").value().toUInt(&ok);
  if ( !ok )
    throw std::runtime_error("invalid from account");

  m_to = element.attributeNode("to").value().toUInt(&ok);
  if ( !ok )
    throw std::runtime_error("invalid to account");

  m_value = element.attributeNode("value").value().toFloat(&ok);
  if ( !ok )
    throw std::runtime_error("invalid value");

  m_date = QDate::fromString(element.attributeNode("date").value(), Qt::ISODate);
  if ( !m_date.isValid() )
    throw std::runtime_error("invalid date");

  m_description = element.attributeNode("description").value();
  
  // oh alright, I'll reduce the number of warnings...
  if ( !element.attributeNode("nextnotify").isNull() )
    m_nextNotify = QDateTime::fromString(element.attributeNode("nextnotify").value(), Qt::ISODate);

  m_recurrence.setRecurStart(m_date);
  QDomNodeList nodes = element.elementsByTagName("recurrence");
  if ( nodes.count() )
    readRecurrence(nodes.item(0).toElement());
}

Transaction::Transaction(KBudgetDoc* d, int i, int from, int to, float val, QDate dt, QString desc)
  : m_doc(d), m_id(i), m_from(from), m_to(to), m_value(val), m_date(dt),
  m_description(desc), m_recurrence(0)
{
}

Transaction::~Transaction() {
}

void Transaction::serialize(QDomElement &parent)
{
  QDomElement e = parent.ownerDocument().createElement("transaction");
  e.setAttribute("id", m_id);
  e.setAttribute("from", m_from);
  e.setAttribute("to", m_to);
  e.setAttribute("value", m_value);
  e.setAttribute("date", m_date.toString(Qt::ISODate));
  e.setAttribute("description", m_description);
  if ( m_nextNotify.isValid() )
    e.setAttribute("nextnotify", m_nextNotify.toString(Qt::ISODate));
  parent.appendChild( e );

  if ( m_recurrence.doesRecur() )
    writeRecurrence(e);
}

int Transaction::id() const
{
  return m_id;
}

int Transaction::from() const
{
  return m_from;
}

int Transaction::to() const
{
  return m_to;
}

float Transaction::value() const
{
  return m_value;
}

QDate Transaction::date() const
{
  return m_date;
}

QString Transaction::description() const
{
  return m_description;
}

ushort Transaction::doesRecur() const
{
  return m_recurrence.doesRecur();
}

KCal::Recurrence& Transaction::recurrence()
{
  return m_recurrence;
}

QDate Transaction::nextRecurrence() const
{
  if ( m_recurrence.doesRecur() )
    return m_recurrence.getNextDate(m_date);
  else
    return QDate();
}

QDateTime Transaction::nextNotify() const
{
  if ( m_nextNotify.isValid() )
    return m_nextNotify;
  else
    return QDateTime(m_recurrence.getNextDate(m_date), QTime(0,0,0));
}

void Transaction::setNextNotify(QDateTime next)
{
  m_nextNotify = next;
  // ok, so we don't really need all the stuff this does... but the next notify
  // *is* part of the transaction data, and it has been changed
  m_doc->signalTransactionChanged(m_id);
}

void Transaction::setData(int from, int to, float val, QDate d, QString desc)
{
  if ( m_doc->canAddUndo() )
    m_doc->addUndo(new KBTransactionUndo(m_doc, KBTransactionUndo::UT_MODIFIED, m_id));

  if ( m_from >= 0 )
    m_doc->getAccount(m_from).removeTransaction(m_id);
  if ( m_to >= 0 )
    m_doc->getAccount(m_to).removeTransaction(m_id);
  m_from = from;
  m_to = to;
  m_value = val;
  m_date = d;
  m_description = desc;
  m_recurrence.setRecurStart(m_date);
  m_doc->getAccount(m_from).addTransaction(m_id);
  m_doc->getAccount(m_to).addTransaction(m_id);

  m_doc->signalTransactionChanged(m_id);
  m_doc->signalTransactionRecurrenceChanged(m_id);
}

void Transaction::setFrom(int id)
{
  if ( m_doc->canAddUndo() )
    m_doc->addUndo(new KBTransactionUndo(m_doc, KBTransactionUndo::UT_MODIFIED, m_id));

  if ( m_from >= 0 )
    m_doc->getAccount(m_from).removeTransaction(m_id);
  m_from = id;
  if ( m_from >= 0 )
    m_doc->getAccount(m_from).addTransaction(m_id);

  m_doc->signalTransactionChanged(m_id);
}

void Transaction::setTo(int id)
{
  if ( m_doc->canAddUndo() )
    m_doc->addUndo(new KBTransactionUndo(m_doc, KBTransactionUndo::UT_MODIFIED, m_id));

  if ( m_to >= 0 )
    m_doc->getAccount(m_to).removeTransaction(m_id);
  m_to = id;
  if ( m_to >= 0 )
    m_doc->getAccount(m_to).addTransaction(m_id);

  m_doc->signalTransactionChanged(m_id);
}

void Transaction::setDescription(QString description)
{
  if ( m_doc->canAddUndo() )
    m_doc->addUndo(new KBTransactionUndo(m_doc, KBTransactionUndo::UT_DESCRIPTION, m_id));
  m_description = description;

  m_doc->signalTransactionChanged(m_id);
}

void Transaction::recurrenceUpdated()
{
  // the next notification time may not be valid anymore
  m_nextNotify = QDateTime();
  m_doc->signalTransactionRecurrenceChanged(m_id);
}

void Transaction::readRecurrence(QDomElement element)
{
  KCal::Recurrence *r = &m_recurrence;

  // clear out any old settings;
  r->unsetRecurs();

  int duration = element.attributeNode("duration").value().toInt();
  int frequency = element.attributeNode("frequency").value().toInt();
  QDate endDate;
  if ( duration == 0 )
    endDate = QDate::fromString(element.attributeNode("enddate").value(), Qt::ISODate);

  QString type = element.attributeNode("type").value();

  QDomNodeList nodes;
  if ( type == "daily" ) {
      if ( duration != 0 ) r->setDaily( frequency, duration );
      else  r->setDaily( frequency, endDate );
  } else if ( type == "weekly" ) {
      QBitArray days(7);
      nodes = element.elementsByTagName("days");
      if ( nodes.count() )
        for ( uint d = 0; d < 7; d++ )
          if ( !nodes.item(0).toElement().attributeNode(QString("d%1").arg(d)).isNull() )
            days.setBit(d);
      if ( duration != 0 ) r->setWeekly( frequency, days, duration );
      else r->setWeekly( frequency, days, endDate );
  } else if ( type == "monthly" ) {
      uint day = element.attributeNode("day").value().toUInt();
      if ( !element.attributeNode("count").isNull() ) {
          int pos = element.attributeNode("count").value().toInt();

          QBitArray days( 7 );
          days.fill( false );
          days.setBit( day );
          
          if ( duration != 0 )
              r->setMonthly( KCal::Recurrence::rMonthlyPos, frequency, duration );
          else
              r->setMonthly( KCal::Recurrence::rMonthlyPos, frequency, endDate );
          r->addMonthlyPos( pos, days );
          if ( r->monthPositions().isEmpty() ) {
            kdDebug() << "Warning: monthly recurrence lacks position attibutes" << endl;
            r->addMonthlyPos( m_date.day()/7 + 1, m_date.dayOfWeek() );
          }
      } else {
          // it's by day
          if ( duration != 0 ) {
              r->setMonthly( KCal::Recurrence::rMonthlyDay, frequency, duration );
          } else {
              r->setMonthly( KCal::Recurrence::rMonthlyDay, frequency, endDate );
          }
          r->addMonthlyDay( day );
          if ( r->monthDays().isEmpty() ) {
            kdDebug() << "Warning: monthly recurrence lacks day attribute" << endl;
            r->addMonthlyDay(m_date.day());
          }
      }
  } else if ( type == "yearly" ) {
    if ( duration != 0 ) {
      r->setYearly( KCal::Recurrence::rYearlyMonth, frequency, duration );
    } else {
      r->setYearly( KCal::Recurrence::rYearlyMonth, frequency, endDate );
    }

    r->addYearlyNum(element.attributeNode("month").value().toInt());
    if ( r->yearNums().isEmpty() ) {
      kdDebug() << "Warning: yearly recurrence lacks month attribute, using " << m_date.month() << endl;
      r->addYearlyNum(m_date.month());
    }
  }

  DateList exceptions;
  QDomNodeList exs = element.elementsByTagName("exception");
  for ( uint i = 0; i < exs.count(); i++ )
    exceptions.append(QDate::fromString(exs.item(i).toElement().text(), Qt::ISODate));
  r->setExDates( exceptions );
}

void Transaction::writeRecurrence(QDomElement &parent)
{
  QValueList<KCal::Recurrence::rMonthPos> rmp;
  QValueList<int> rmd;
  int day = 0;
  int count = 0;

  int recurs = m_recurrence.doesRecur();

  QDomElement e = parent.ownerDocument().createElement("recurrence");
  parent.appendChild(e);
  e.setAttribute("frequency", m_recurrence.frequency());
  QDomElement e2;
  switch ( recurs ) {
  case KCal::Recurrence::rDaily:
    e.setAttribute("type", "daily");
    break;
  case KCal::Recurrence::rWeekly:
    e.setAttribute("type", "weekly");
    e2 = e.ownerDocument().createElement("days");
    e.appendChild(e2);
    for ( uint i = 0; i < m_recurrence.days().size(); i++ )
      if ( m_recurrence.days().testBit(i) )
        e2.setAttribute(QString("d%1").arg(i), 1);
    break;
  case KCal::Recurrence::rMonthlyPos:
    e.setAttribute("type", "monthly");

    rmp = m_recurrence.monthPositions();
    if ( rmp.first().negative )
      count=-rmp.first().rPos;
    else
      // give the week as -5 to -1 and 1 to 5. the widget will do the rest
      count = rmp.first().rPos;
    day = 0;
    while ( !rmp.first().rDays.testBit( day ) ) ++day;
    e.setAttribute("count", count);
    e.setAttribute("day", day);

    break;
  case KCal::Recurrence::rMonthlyDay:
    e.setAttribute("type", "monthly");

    rmd = m_recurrence.monthDays();
    day = !rmd.isEmpty() ? rmd.first() : date().day();
    e.setAttribute("day", day);

    break;
  case KCal::Recurrence::rYearlyMonth:
  case KCal::Recurrence::rYearlyDay:
    e.setAttribute("type", "yearly");
    rmd = m_recurrence.yearNums();
    e.setAttribute("month", rmd.first());
    break;
  default:
    break;
  }
  
  e.setAttribute("duration", m_recurrence.duration());
  if ( m_recurrence.duration() == 0 )
    e.setAttribute("enddate", m_recurrence.endDate().toString(Qt::ISODate));

  DateList dates = m_recurrence.exDates();
  for ( uint i = 0; i < dates.count(); i++ ) {
    QDomElement ex = e.ownerDocument().createElement("exception");
    e.appendChild(ex);
    QDomText extxt = ex.ownerDocument().createTextNode(dates[i].toString(Qt::ISODate));
    ex.appendChild(extxt);
  }
    
}
