/***************************************************************************
                          transaction.h  -  description
                             -------------------
    begin                : Sat Aug 10 2002
    copyright            : (C) 2002 by Richard Garand
    email                : richard@garandnet.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TRANSACTION_H
#define TRANSACTION_H

#include <qdatetime.h>
#include <qstring.h>
#include <qdom.h>

#include <libkcal/recurrence.h>

class KBudgetDoc;
class QTextStream;

/**A class that describes a transaction between two accounts
  *@author Richard Garand
  */

class Transaction {
  public:
    /** default constructor (if doc == 0, no other functions can be called) */
    Transaction(KBudgetDoc* doc = 0);
    /** read data from a stream; version is the stream's format version */
    Transaction(KBudgetDoc* doc, int id, QTextStream &stream, int version);
    /** read data from an XML element */
    Transaction(KBudgetDoc* doc, int id, QDomElement element);
    /** initialize a transaction object */
    Transaction(KBudgetDoc* doc, int id, int from, int to, float value, QDate date, QString description);
    /** default destructor */
    ~Transaction();

    /** Writes the data to an XML document as an element under parent. */
    void serialize(QDomElement &parent);
    
    /** return the ID of the transactions */
    int id() const;
    /** return the ID of the account this is from */
    int from() const;
    /** return the ID of the account this is to */
    int to() const;
    /** get the value */
    float value() const;
    /** get the date */
    QDate date() const;
    /** get the description */
    QString description() const;
    
    /** Returns whether the transaction recurs. */
    ushort doesRecur() const;
    /** Returns the recurrence for the transaction. If the return value is
      * modified, recurrenceUpdated() must be called. */
    KCal::Recurrence& recurrence();
    /** Returns the date of the next recurrence. The return value is undefined
      * for a non-recurring transaction. */
    QDate nextRecurrence() const;
    
    /** Returns the date and time when the user should be notified about the
      * next recurrence. */
    QDateTime nextNotify() const;
    /** Sets the date and time when the user should be notified about the next
      * recurrence. */
    void setNextNotify(QDateTime next);

    /** Set the accounts, value, date, and description. This creates an undo
      * action if possible, and includes the recurrence data; using this
      * function is currently the only way to save the recurrence data for an
      * undo action. */
    void setData(int from, int to, float value, QDate date, QString description);
    /** Sets the from account; -1 for no account. */
    void setFrom(int id);
    /** Sets the to account; -1 for no account. */
    void setTo(int id);
    /** Set the description, creating an undo action if possible. */
    void setDescription(QString description);

    /** Signals that the returrence has been changed. */
    void recurrenceUpdated();
  private:
    KBudgetDoc* m_doc;
    int m_id;
    int m_from;
    int m_to;
    float m_value;
    QDate m_date;
    QString m_description;

    KCal::Recurrence m_recurrence;
    QDateTime m_nextNotify;

    /** Reads recurrence data from a DOM recurrence element. */
    void readRecurrence(QDomElement element);
    /** Creates and fills a recurrence element under parent. */
    void writeRecurrence(QDomElement &parent);
};

#endif
